#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "fire-sparse.h"
#include "memutils.h"
#include "types.h"

int32 checkValueI16( int16 val, int16 b1, int16 b2 )
{
  int32 ret;

  ret = (val >= b1) && (val <= b2);
  if (!ret) fprintf( stderr,
		     "Wrong parameter: %d <= %d <= %d not satisfied!\n",
		     b1, val, b2 );
  return( ret );
}

int32 checkValueI32( int32 val, int32 b1, int32 b2 )
{
  int32 ret;

  ret = (val >= b1) && (val <= b2);
  if (!ret) fprintf( stderr,
		     "Wrong parameter: %d <= %d <= %d not satisfied!\n",
		     b1, val, b2 );
  return( ret );
}

int32 checkValueF( float val, float b1, float b2 )
{
  int32 ret;

  ret = (val >= b1) && (val <= b2);
  if (!ret) fprintf( stderr,
		     "Wrong parameter: %f <= %f <= %f not satisfied!\n",
		     b1, val, b2 );
  return( ret );
}

int32 fsLoadParameters( FParams *par )
{
  fprintf( stdout, "Frames per second: " );
  fscanf( stdin, SCANI16, &par->fPS );
  fprintf( stdout, "%d\n", par->fPS );
  if (!checkValueI16( par->fPS, 1, 100 )) goto error;

  fprintf( stdout, "Fire width: " );
  fscanf( stdin, SCANI16, &par->fireW );
  fprintf( stdout, "%d\n", par->fireW );
  if (!checkValueI16( par->fireW, 30, 2048 )) goto error;

  fprintf( stdout, "Fire height: " );
  fscanf( stdin, SCANI16, &par->fireH );
  fprintf( stdout, "%d\n", par->fireH );
  if (!checkValueI16( par->fireH, 30, 2048 )) goto error;

  fprintf( stdout, "Max. power: " );
  fscanf( stdin, SCANI16, &par->maxPower );
  fprintf( stdout, "%d\n", par->maxPower );
  if (!checkValueI16( par->maxPower, 1, 255 )) goto error;

  fprintf( stdout, "Max. FPixels: " );
  fscanf( stdin, SCANI32, &par->maxFPixels );
  fprintf( stdout, "%d\n", par->maxFPixels );
  if (!checkValueI32( par->maxFPixels, 1, 1000000 )) goto error;

  fprintf( stdout, "Max. random numbers: " );
  fscanf( stdin, SCANI16, &par->maxRandom );
  fprintf( stdout, "%d\n", par->maxRandom );
  if (!checkValueI16( par->maxRandom, 1, 1000 )) goto error;

  fprintf( stdout, "Random range: " );
  fscanf( stdin, SCANI16, &par->randomRange );
  fprintf( stdout, "%d\n", par->randomRange );
  if (!checkValueI16( par->randomRange, 1, par->maxPower )) goto error;

  fprintf( stdout, "Random decrease: " );
  fscanf( stdin, SCANI16, &par->randomDecrease );
  fprintf( stdout, "%d\n", par->randomDecrease );
  if (!checkValueI16( par->randomDecrease, -par->maxPower, par->maxPower ))
      goto error;

  fprintf( stdout, "Power to survive: " );
  fscanf( stdin, SCANI16, &par->powerSurvive );
  fprintf( stdout, "%d\n", par->powerSurvive );
  if (!checkValueI16( par->powerSurvive, 0, par->maxPower )) goto error;

  fprintf( stdout, "Power to create: " );
  fscanf( stdin, SCANI16, &par->powerCreate );
  fprintf( stdout, "%d\n", par->powerCreate );
  if (!checkValueI16( par->powerCreate, 0, par->maxPower )) goto error;

  fprintf( stdout, "Power of overheating: " );
  fscanf( stdin, SCANI16, &par->powerOverheat );
  fprintf( stdout, "%d\n", par->powerOverheat );
  if (!checkValueI16( par->powerOverheat, 0, 8 * par->maxPower )) goto error;

  fprintf( stdout, "Power reduction due to overheat: " );
  fscanf( stdin, "%f", &par->powerReduction );
  fprintf( stdout, "%f\n", par->powerReduction );
  if (!checkValueF( par->powerReduction, 0.0, 1.0 )) goto error;

/*   par->fPS = 30; */
/*   par->fireW = 300; */
/*   par->fireH = 300; */
/*   par->maxColors = 256; */
/*   par->maxPower = 255; */
/*   par->maxFPixels = 10000; */
/*   par->maxRandom = 100; */
/*   par->randomRange = 70; */
/*   par->randomDecrease = 46; */
/*   par->powerSurvive = 0; */
/*   par->powerCreate = 23; */
/*   par->powerOverheat = 2040; */
/*   par->powerReduction = 0.5; */

  par->maxColors = 256;

  return( 0 );

 error:
  return( 1 );
}

int32 fsCreateData( FParams *par, FSparseData *data )
{
  if (data) {
    (data)->nFP = 0;
    (data)->bufFP =
      (uchar *) allocMem( par->fireW * par->fireH * sizeof(uchar) );
    (data)->listFP = (FPixel *) allocMem( par->maxFPixels * sizeof(FPixel) );
  } else return( 1 );
  return( 0 );
}

int32 fsClearData( FParams *par, FSparseData *data )
{
  data->nFP = 0;
  memset( data->bufFP, 0, par->fireW * par->fireH * sizeof(uchar) );
  return( 0 );
}

int32 fsDestroyData( FSparseData *data )
{
  if (data) {
    if (data->bufFP) freeMem( data->bufFP );
    data->bufFP = 0;
    if (data->listFP) freeMem( data->listFP );
    data->listFP = 0;
    data->nFP = 0;
  }
  return( 0 );
}

int32 fsClearDataFast( FParams *par, FSparseData *data, GridInfo *gi )
{
  int32 i, j, x, y, w, h;
  uchar *p;

  data->nFP = 0;
  for (i = 0; i < (gi->nx * gi->ny); i++) {
    if (gi->mask0[i]) {
      x = gi->x[i%gi->nx];
      w = gi->w[i%gi->nx];
      y = gi->y[i/gi->nx];
      h = gi->h[i/gi->nx];
      p = data->bufFP + y * par->fireW + x;
      for (j = 0; j < h; j++) {
	memset( p + j * par->fireW, 0, w * sizeof( uchar ) );
      }
    }
  }
  return( 0 );
}

void addFPixel( uchar *buf, FPixel *list, uint32 *num, uint32 max,
		uint32 offset, int32 power )
{
  if (*num < max) {
    list[*num].offset = offset;
    list[*num].power = power;
    buf[offset] = (uchar) power;
    //    printf( "%d\n", buf[offset] );
    (*num)++;
  }
}

int32 getPowerAround( uchar *buf, uint32 w, uint32 of )
{
  int power;
  uchar *p;

  /*  
  power = *(buf + of - w - 1);
  power += *(buf + of - w);
  power += *(buf + of - w + 1);
  power += *(buf + of - 1);
  power += *(buf + of + 1);
  power += *(buf + of + w - 1);
  power += *(buf + of + w);
  power += *(buf + of + w + 1);
  */
  
  p = buf + of - w - 1;
  power = *(p);
  power += *(p+1);
  power += *(p+2);
  p += w;
  power += *(p);
  power += *(p+2);
  p += w;
  power += *(p);
  power += *(p+1);
  power += *(p+2);

  return( power );
}


int32 fsInit( Filter *f, FParams *par )
{
  int32 i;
  FParams *p;

  if (f->type != FSFilter) return( 1 );

  f->data = allocMem( sizeof(FParams) );
  p = f->data;

  p->fPS = par->fPS;
  p->fireW = par->fireW;
  p->fireH = par->fireH;
  p->maxColors = par->maxColors;
  p->maxPower = par->maxPower;
  p->maxFPixels = par->maxFPixels;
  p->maxRandom = par->maxRandom;
  p->randomRange = par->randomRange;
  p->randomDecrease = par->randomDecrease;
  p->powerSurvive = par->powerSurvive;
  p->powerCreate = par->powerCreate;
  p->powerOverheat = par->powerOverheat;
  p->powerReduction = par->powerReduction;

/*    printf( "%d\n", RAND_MAX ); */
  p->randNum = (uint16 *) allocMem( p->maxRandom * sizeof(uint16) );
  srand( (uint) time(0) );
  for (i = 0; i < p->maxRandom; i++) {
    p->randNum[i] = rand() % p->randomRange - p->randomDecrease;
  }
  p->rc = 0;

  p->xy[0] = 0;
  p->xy[1] = 1;
  p->xy[2] = 1;
  p->xy[3] = p->fireW - 2;
  p->xy[4] = 2;
  p->xy[5] = p->fireW - 2;
  p->xy[6] = 1;
  p->xy[7] = 1;
  return( 0 );
}

int32 fsClear( Filter *f )
{
  FParams *p;

  if (f->type != FSFilter) return( 1 );
  p = f->data;
  if (p) {
    freeMem( p->randNum );
    freeMem( p );
    f->data = 0;
  } 
  return( 0 );
}

int32 fsAction( Filter *f, FSparseData *in, FSparseData *out )
{
  int32 i, j, power, of, pow, stab;
  int32 x, y;
  FParams *p;

  if (f->type != FSFilter) return( 1 );
  p = f->data;


  stab = 1;
  out->nFP = 0;
/*   printf( "%d %d\n", in->nFP, out->nFP ); */
  for (i = 0; i < in->nFP; i++) {
    x = in->listFP[i].offset % p->fireW;
    y = in->listFP[i].offset / p->fireW;
    if ((x <= 1) || (x >= (p->fireW-1)) || (y <= 1) || (y >= (p->fireH-1)))
      continue;
    power = 0;
    of = in->listFP[i].offset - p->fireW - 1;
    for (j = 0; j < 8; j++, of += p->xy[j]) {
      if (in->bufFP[of] != 0) {
	  power += in->bufFP[of];
      } else {
	if (out->bufFP[of] == 0) {
	  p->rc = (p->rc + 1) % p->maxRandom;
	  pow = (getPowerAround( in->bufFP, p->fireW, of ) >> 3) +
	    p->randNum[p->rc];
/* 	  printf( "%d %d %d\n", pow, p->rc, p->randNum[p->rc] ); */
	  if (pow > p->powerCreate ) {
	    if (pow > p->maxPower) pow = p->maxPower;
	    
	    addFPixel( out->bufFP, out->listFP, &out->nFP,
		       p->maxFPixels, of, pow );
	    /*
	    if (out->nFP < p->maxFPixels) {
	      out->listFP[out->nFP].offset = of;
	      out->listFP[out->nFP].power = pow;
	      out->bufFP[of] = (uchar) pow;
	      (out->nFP)++;
	    }
	    */
	  }
	}
      }
    }
    if (power >= p->powerOverheat) {
      power = (int32) (p->powerReduction * power);
    }
    of = in->listFP[i].offset;
    power += in->bufFP[of];
    p->rc = (p->rc + 1) % p->maxRandom;
    power = (power >> 3) + p->randNum[p->rc];
    if (power > p->powerSurvive ) {
      if (!(stab = (stab + 1) % 4)) {
	if (power > p->maxPower * 0.8) power *= 0.9;
	else if (power < p->maxPower * 0.3) power *= 1.2;
      }
      if (power > p->maxPower) power = p->maxPower;

      addFPixel( out->bufFP, out->listFP, &out->nFP,
		 p->maxFPixels, of, power );
      /*
      if (out->nFP < p->maxFPixels) {
	out->listFP[out->nFP].offset = of;
	out->listFP[out->nFP].power = power;
	out->bufFP[of] = (uchar) power;
	(out->nFP)++;
      }
      */
    }
  }
  return( 0 );
}
