#ifndef _FIRE_SPARSE_H_
#define _FIRE_SPARSE_H_

#include "types.h"
#include "timer.h"

#define FPointer (void *)
#define FSFilter 0

typedef struct {
  int32 offset;
  uchar power;
} FPixel;

typedef struct {
  uint16 fPS;
  uint16 fireW;
  uint16 fireH;
  uint16 maxColors;
  uint16 maxPower;
  uint32 maxFPixels;
  uint16 maxRandom;
  uint16 randomRange;
  int16 randomDecrease;
  uint16 powerSurvive;
  uint16 powerCreate;
  uint16 powerOverheat;
  float32 powerReduction;

  int16 *randNum;
  uint32 rc;
  int32 xy[8];
} FParams;

typedef struct {
  uint32 totalFrames;
  uint32 frameTag;
  Timer timeStat;
  float32 timeTag;
} TimeInfo;

typedef struct {
  uchar *bufFP;
  FPixel *listFP;
  uint32 nFP;
} FSparseData;

typedef struct filter {
  int32 (*init)( struct filter *f, void *data );
  int32 (*action)( struct filter *f, void *in, void *out );
  int32 (*clear)( struct filter *f );
  int32 type;
  void *data;
} Filter;

typedef struct {
  int32 nx;
  int32 ny;
  int32 *w;
  int32 *h;
  int32 *x;
  int32 *y;
  int32 *xmap;
  int32 *ymap;
  uchar *mask;
  uchar *mask0;
} GridInfo;

int32 fsLoadParameters( FParams *par );
int32 fsCreateData( FParams *par, FSparseData *data );
int32 fsClearData( FParams *par, FSparseData *data );
int32 fsDestroyData( FSparseData *data );
int32 fsClearDataFast( FParams *par, FSparseData *data, GridInfo *gi );
void addFPixel( uchar *buf, FPixel *list, uint32 *num, uint32 max,
		uint32 offset, int32 power );
int32 getPowerAround( uchar *buf, uint32 w, uint32 of );
int32 fsInit( Filter *f, FParams *par );
int32 fsClear( Filter *f );
int32 fsAction( Filter *f, FSparseData *in, FSparseData *out );

#endif
