#include "types.h"
#include "handle.h"
#include "memutils.h"
#include "graphics.h"

#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>

#define RGB32( r, g, b ) (((r) << 16) | ((g) << 8) | b)
//#define RGB32( r, g, b ) ((((r) << 8)&0xf800) | (((g) << 3)&0x7e0) | ((b)&0x1f))
#define READ32(adr) *((int32 *)adr)
#define WRITE32(adr,s) *((int32 *)adr)=s
#define WRITE16( adr, s ) (*((int16 *)adr) = s)

/* From vroot.h by Andreas Stolcke. */
int VirtualRootWindowOfScreen( screenP )
     Screen* screenP;
{
  static Window root = (Window) 0;
  Display* dpy = DisplayOfScreen( screenP );
  Atom __SWM_VROOT = None;
  int i;
  Window rootReturn, parentReturn;
  Window* children;
  unsigned int numChildren;

  root = RootWindowOfScreen( screenP );

  /* Go look for a virtual root. */
  __SWM_VROOT = XInternAtom( dpy, "__SWM_VROOT", False );
  if ( XQueryTree(
                  dpy, root, &rootReturn, &parentReturn, &children,
                  &numChildren ) ) {
    for ( i = 0; i < numChildren; ++i ) {
      Atom actual_type;
      int actual_format;
      unsigned long nitems, bytesafter;
      Window* newRoot = (Window*) 0;

      if ( XGetWindowProperty( dpy, children[i], __SWM_VROOT, 0, 1,
                               False, XA_WINDOW,
                               &actual_type, &actual_format, &nitems,
                               &bytesafter,
                               (unsigned char**) &newRoot ) == Success
           && newRoot ) {
        root = *newRoot;
        break;
      }
    }
    if ( children )
      XFree( (char*) children );
  }

  return root;
}

int topBit( uint32 mask )
{
  int i;

  if (!mask) return 0;

  i = 32;
  while(!(mask&0x80000000))
  {
    mask <<= 1;
    i--;
  }

  return( i );
}

void pal2pal( uint32 *pal1, uint32 *pal2, GraphicsData *gr )
{
  uint32 i;
  uint32 r, g, b;
  uchar rs, gs, bs;

  rs = 24 - topBit( gr->rmask );
  gs = 16 - topBit( gr->gmask );
  bs = 8 - topBit( gr->bmask );
  printf( "Color shifts: %d %d %d\n", rs, gs, bs );

  for (i = 0; i < 256; i++) {
    r = (pal1[i] >> rs) & gr->rmask;
    g = (pal1[i] >> gs) & gr->gmask;
    b = (pal1[i] >> bs) & gr->bmask;
    pal2[i] = r | g | b;
    /*    printf( "%d %x %x\n", i, pal1[i], pal2[i] );*/
  }
}

/*
 * Hermes library 'inspired' code...
 */
void convert8_32( char *source,char *dest,
		  uint32 *palette, uint32 count )
{ int i;
  uchar s_pixel;
  int32 d_pixel;

  for(i=0;i<count;i++,source++,dest+=4)
  { s_pixel=(uchar)*(source);

    d_pixel=palette[s_pixel];
    WRITE32(dest,d_pixel);
  }
}

void convert8_32_masked( char *source,char *dest,
			 uint32 *palette, uint32 count )
{
  int i;
  uchar s_pixel;
  int32 d_pixel;

  for(i=0;i<count;i++,source++,dest+=4) {
    s_pixel=(uchar)*(source);
    if (s_pixel) {
      d_pixel=palette[s_pixel];
      WRITE32(dest,d_pixel);
    }
  }
}

void convert8_24_L( char *source,char *dest,
		  uint32 *palette, uint32 count )
{ int i;
  uchar s_pixel;
  int32 d_pixel;
  uchar *d_ptr=(uchar *)&d_pixel;

  for(i=0;i<count;i++,source++,dest+=3)
  { s_pixel=(uchar)*(source);

    d_pixel=palette[s_pixel];

    *((uchar *)dest+0)=*(d_ptr+0);
    *((uchar *)dest+1)=*(d_ptr+1);
    *((uchar *)dest+2)=*(d_ptr+2);

    WRITE32(dest,d_pixel);
  }
}

void convert8_24_L_masked( char *source,char *dest,
			   uint32 *palette, uint32 count )
{
  int i;
  uchar s_pixel;
  int32 d_pixel;
  uchar *d_ptr=(uchar *)&d_pixel;

  for(i=0;i<count;i++,source++,dest+=3) {
    s_pixel=(uchar)*(source);
    if (s_pixel) {
      d_pixel=palette[s_pixel];
      
      *((uchar *)dest+0)=*(d_ptr+0);
      *((uchar *)dest+1)=*(d_ptr+1);
      *((uchar *)dest+2)=*(d_ptr+2);
      
      WRITE32(dest,d_pixel);
    }
  }
}

void convert8_24_B( char *source,char *dest,
		  uint32 *palette, uint32 count )
{ int i;
  uchar s_pixel;
  int32 d_pixel;
  uchar *d_ptr=(uchar *)&d_pixel;

  for(i=0;i<count;i++,source++,dest+=3)
  { s_pixel=(uchar)*(source);

    d_pixel=palette[s_pixel];

    *((uchar *)dest+0)=*(d_ptr+3);
    *((uchar *)dest+1)=*(d_ptr+2);
    *((uchar *)dest+2)=*(d_ptr+1);

    WRITE32(dest,d_pixel);
  }
}

void convert8_24_B_masked( char *source,char *dest,
			   uint32 *palette, uint32 count )
{
  int i;
  uchar s_pixel;
  int32 d_pixel;
  uchar *d_ptr=(uchar *)&d_pixel;

  for(i=0;i<count;i++,source++,dest+=3) {
    s_pixel=(uchar)*(source);
    if (s_pixel) {
      d_pixel=palette[s_pixel];
      
      *((uchar *)dest+0)=*(d_ptr+3);
      *((uchar *)dest+1)=*(d_ptr+2);
      *((uchar *)dest+2)=*(d_ptr+1);
      
      WRITE32(dest,d_pixel);
    }
  }
}
 
void convert8_16( char *source, char *dest, 
		  uint32 *palette, uint32 count )
{ int i;
  uchar s_pixel;
  int32 d_pixel;
  
  if (((int)dest&0x3)!=0)
  { s_pixel=(uchar)*(source);
    WRITE16(dest,palette[s_pixel]);
    source++; 
    dest+=2;
    count--;
  }

  for(i=0;i<count>>1;i++,source+=2,dest+=4)
  { 
    d_pixel=palette[(uchar)*(source)];
    d_pixel|=palette[(uchar)*(source+1)]<<16;

    WRITE32(dest,d_pixel);
  }
  s_pixel = (uchar)*(source);
  /*  printf( "%x, %d\n", palette[s_pixel], s_pixel );*/

  if(count&1)
  { s_pixel=(uchar)*(source);
   
    WRITE16(dest,palette[s_pixel]);
  }
}

void convert8_16_masked( char *source, char *dest, 
			 uint32 *palette, uint32 count )
{
  int i;
  uchar s_pixel;
  
  for (i = 0; i < count; i++, source++, dest += 2) {
    s_pixel = (uchar)*(source);
    if (s_pixel) WRITE16( dest, palette[s_pixel] );
  }
}

void convert8_i8( char *source, char *dest,
		  uint32 *palette, uint32 count )
{ int i;
  int32 s_block;
  uchar s_pixel;

  for(i=0;i<count>>2;i++,source+=4,dest+=4)
  { s_block=READ32(source);

    WRITE32(dest,s_block);
  }

  for(i=0;i<(count&3);i++,source++,dest++)
  { s_pixel=*((uchar *)source);

    *((uchar *)dest)=s_pixel;
  }
}

void convert8_i8_masked( char *source, char *dest,
			 uint32 *palette, uint32 count )
{ int i;
  uchar s_pixel;

  for(i=0;i<count;i++,source+=1,dest+=1) {
    s_pixel=*((uchar *)source);
    if (s_pixel) {
      *((uchar *)dest)=s_pixel;
    }
  }
}

/*
 * 04.09.2000, 0.0.12, c
 */
void generateStdColormap( uint *colors )
{
  int i;

  for (i = 0; i < 256; i++) {
    colors[i] = RGB32( 255, 255, 255 );
  }

  colors[0] = RGB32( 0, 0, 0 );
  for (i = 1; i < 85; i++) {
    colors[i] = RGB32( 85 + 2 * i, i, 0 );
  }
  for (i = 0; i < 85; i++) {
    colors[85+i] = RGB32( 255, 85 + 2 * i, 0 );
  }
  for (i = 0; i < 85; i++) {
    colors[170+i] = RGB32( 255, 255, 3 * i );
  }
  colors[255] = RGB32( 255, 255, 255 );
}

/*
 * 04.09.2000, 0.0.12
 * 07.09.2000, 0.0.14
 */
int32 initGraphics( GraphicsData *p, char *name, uint16 *w, uint16 *h )
{
  int i, n, maxColors = 256;
  int aux, tmp[3];
  FILE *file;
  Screen *screen;
  Visual *visual;
  Cursor cursor;
  Pixmap source, mask;
  XColor foreCol, backCol;
  XPixmapFormatValues *pixFormats;
  GC cursorGC;
  int bitmapPad, pixFound;
   
  printf( "Initializing graphics...\n" );

  p->thedisplay  = XOpenDisplay(NULL);
  p->blackColor  = BlackPixel(p->thedisplay,DefaultScreen(p->thedisplay));
  p->whiteColor  = WhitePixel(p->thedisplay,DefaultScreen(p->thedisplay));
  p->thescreen   = DefaultScreen(p->thedisplay);
  p->thecolormap = DefaultColormap(p->thedisplay,p->thescreen);

  if (p->inRootWindow) {
    screen = ScreenOfDisplay( p->thedisplay, p->thescreen );
    p->w = WidthOfScreen( screen );
    p->h = HeightOfScreen( screen );
    *w = p->w;
    *h = p->h;
    p->thewindow = VirtualRootWindowOfScreen( screen );
    /*    p->thewindow = DefaultRootWindow(p->thedisplay); */
  } else {
    p->w = *w;
    p->h = *h;

/*      XAutoRepeatOff( p->thedisplay ); */
    
    /* Create the window */
    p->thewindow = XCreateSimpleWindow(p->thedisplay, 
				       DefaultRootWindow(p->thedisplay),0,0, 
				       p->w,p->h,0,p->blackColor,p->blackColor);
    XMapWindow(p->thedisplay,p->thewindow);
    
    XStringListToTextProperty(&name,1,&p->textproperty);
    XSetWMName(p->thedisplay,p->thewindow,&p->textproperty);
  }



  XSelectInput(p->thedisplay,p->thewindow,StructureNotifyMask | KeyPressMask
	       | PointerMotionMask );

  if (!p->inRootWindow) {
    /* Wait for the MapNotify event */
    for (;;) {
      XNextEvent(p->thedisplay, &p->anevent);
      if (p->anevent.type == MapNotify)
	break;
    }
  }

  /* Get the context */
  p->thecontext = XCreateGC(p->thedisplay,p->thewindow,0,NULL);
  XSetBackground(p->thedisplay,p->thecontext,p->blackColor);
  XSetForeground(p->thedisplay,p->thecontext,p->whiteColor);
  XSetFillStyle(p->thedisplay,p->thecontext,FillSolid);
  

  /* Set a cursor. */
/*    cursor = XCreateFontCursor( p->thedisplay, XC_tcross ); */
  source = XCreatePixmap( p->thedisplay, p->thewindow, 11, 11, 1 );
  cursorGC = XCreateGC( p->thedisplay, source, 0, NULL );
  XSetForeground( p->thedisplay, cursorGC, p->blackColor );
  XFillRectangle( p->thedisplay, source, cursorGC, -1, -1, 12, 12 );
  XSetForeground( p->thedisplay, cursorGC, p->whiteColor );
  XDrawArc( p->thedisplay, source, cursorGC, 0, 0, 10, 10, 0, 360*64 );
  mask = source;
  XParseColor( p->thedisplay, p->thecolormap, "white", &foreCol );
  XParseColor( p->thedisplay, p->thecolormap, "black", &backCol );
  cursor = XCreatePixmapCursor( p->thedisplay, source, mask,
				&foreCol, &backCol, 5, 5 );
  XFreePixmap( p->thedisplay, source );
  XDefineCursor( p->thedisplay, p->thewindow, cursor );

  /* Erase the display (In the background colour) */
  XClearWindow(p->thedisplay,p->thewindow);

  XGetGeometry( p->thedisplay, p->thewindow, &p->root,
		&p->xr, &p->yr, &p->wr, &p->hr, &p->br, &p->dr );
  
  printf( "Display: %d %d %d %d %d %d\n",
	  p->xr, p->yr, p->wr, p->hr, p->br, p->dr );

  pixFormats = XListPixmapFormats( p->thedisplay, &n );
  pixFound = 0;
  printf( "Supported pixmaps:\n" );
  for (i = 0; i < n; i++) {
    printf( " depth %d, bits_per_pixel %d, scanline_pad %d\n",
	    pixFormats[i].depth, pixFormats[i].bits_per_pixel,
	    pixFormats[i].scanline_pad );
    if (pixFormats[i].depth == p->dr) {
      p->bitsPerPixel = pixFormats[i].bits_per_pixel;
      pixFound = 1;
      printf( " -> suitable, not searching further.\n" );
      break;
    }
  }
  XFree( pixFormats );
  if (!pixFound) {
    fprintf( stderr, "No pixmap with depth %d found!\n", p->dr );
    exit( 1 );
  }

  bitmapPad = BitmapPad( p->thedisplay );
  printf( "Bitmap pad: %d\n", bitmapPad );
  p->byteOrder = ImageByteOrder( p->thedisplay );
  printf( "Image byte order: %d\n", p->byteOrder );

  visual = DefaultVisual( p->thedisplay,
			  DefaultScreen( p->thedisplay ) );

  printf( "Visual id: %lx\n", visual->visualid );
  printf( "Visual class: %d\n", visual->class );
  printf( "Color masks: %lx %lx %lx\n",
	  visual->red_mask, visual->green_mask, visual->blue_mask );
  printf( "Bits per rgb: %d\n", visual->bits_per_rgb );
  printf( "Colormap entries: %d\n", visual->map_entries );

  if ((visual->class == TrueColor) || (visual->class == DirectColor))
    p->indexed = 0;
  else
    p->indexed = 1;

  switch (p->bitsPerPixel) {
  case 8:
    p->bytesPerPixel = 1;
    if (p->isBackground)
      p->convert = &convert8_i8_masked;
    else
      p->convert = &convert8_i8;
    break;
  case 16:
    p->bytesPerPixel = 2;
    if (p->isBackground)
      p->convert = &convert8_16_masked;
    else
      p->convert = &convert8_16;
    /*    p->convert = &convert8_16;*/
    break;
  case 24:
    p->bytesPerPixel = 3;
    if (p->isBackground) {
      if (p->byteOrder == LSBFirst)
	p->convert = &convert8_24_L_masked;
      else 
	p->convert = &convert8_24_B_masked;
    } else {
      if (p->byteOrder == LSBFirst)
	p->convert = &convert8_24_L;
      else 
	p->convert = &convert8_24_B;
    }
    break;
  case 32:
    p->bytesPerPixel = 4;
    if (p->isBackground)
      p->convert = &convert8_32_masked;
    else
      p->convert = &convert8_32;
    break;
  default:
    fprintf( stderr, "%d bpp not supported (yet)!\n", p->dr );
    exit( 1 );
  }
  if (p->indexed && (p->dr != 8)) {
    fprintf( stderr, "%d bpp indexed mode not supported (yet)!\n", p->dr );
    exit( 1 );
  }

  printf( "Bytes per pixel: %d\n", p->bytesPerPixel );
  
  p->rmask = visual->red_mask;
  p->gmask = visual->green_mask;
  p->bmask = visual->blue_mask;
  
  p->image = (uchar *) allocMem( p->w * p->h *
				 p->bytesPerPixel * sizeof(uchar) );
  p->ximg = XCreateImage( p->thedisplay,
			  DefaultVisual( p->thedisplay,
					 DefaultScreen( p->thedisplay ) ),
			  DefaultDepth( p->thedisplay,
					DefaultScreen( p->thedisplay ) ),
			  ZPixmap, 0, (char *) p->image,
			  p->w, p->h, bitmapPad, p->bytesPerPixel * p->w );
  if (p->isBackground) {
    p->back = XGetImage( p->thedisplay, p->thewindow,
			 p->xr, p->yr, p->wr, p->hr, AllPlanes, ZPixmap );
  }
  /*
   * Allocate fire colors.
   */
  p->colors = (uint *) allocMem( maxColors * sizeof(uint) );
  p->pal = (uint *) allocMem( maxColors * sizeof(uint) );

  /*
   * Try to read colormap.
   */
  if (p->colormapFile) {
    if ((file = fopen( p->colormapFile, "r" ))  != 0) {
      printf( "Using colormap '%s'.\n", p->colormapFile );
      for (i = 0; i < maxColors; i++) {
	fscanf( file, "%d%d%d%d", &aux, tmp, tmp + 1, tmp + 2 );
	p->colors[i] = RGB32( tmp[0], tmp[1], tmp[2] );
      }
      fclose( file );
    } else {
      printf( "Using colormap: 'default'!\n" );
      generateStdColormap( p->colors );
    }
  } else {
    printf( "Using colormap: 'default'!\n" );
    generateStdColormap( p->colors );
  }

  pal2pal( p->colors, p->pal, p );

  p->x0 = 0;
  p->y0 = 0;

  /*
   * Pixmap for printing statistics.
   */
  p->pixInfo = XCreatePixmap( p->thedisplay, p->thewindow, 80, 40, p->dr );
  p->gcInfo = XCreateGC(p->thedisplay,p->pixInfo,0,NULL);
  XSetBackground(p->thedisplay,p->gcInfo,p->blackColor);
  XSetForeground(p->thedisplay,p->gcInfo,p->whiteColor);
  XSetFillStyle(p->thedisplay,p->gcInfo,FillSolid);

  return( 0 );
}

int32 clearGraphics( GraphicsData *p )
{
  freeMem( p->colors );
  freeMem( p->pal );
  /* Done */
  /*   sleep(5);*/
  /*   XDestroyImage( ximg );*/
  if (p->isBackground) {
    XDestroyImage( p->back );
  }
  XFreeGC(p->thedisplay,p->gcInfo);  
  XFreePixmap( p->thedisplay, p->pixInfo );
  XFreeGC(p->thedisplay,p->thecontext);  
  XUndefineCursor( p->thedisplay, p->thewindow );
  if (!p->inRootWindow) {
    XUnmapWindow(p->thedisplay,p->thewindow);
    XDestroyWindow(p->thedisplay,p->thewindow);
/*      XAutoRepeatOn( p->thedisplay ); */
  }
  XFlush( p->thedisplay );
  XCloseDisplay(p->thedisplay);

  return( 0 );
}

int32 processEvents( GraphicsData *p )
{
  int32 isevent, keycode;

  while ((isevent = XPending( p->thedisplay ))) {
    XNextEvent( p->thedisplay, &p->anevent );
/*      printf( "%d %d\n", isevent, p->anevent.type ); */
    switch (p->anevent.type) {
      
    case KeyPress:
      keycode = ((XKeyEvent *) &p->anevent)->keycode;
      p->keysym = XLookupKeysym( (XKeyPressedEvent *) &p->anevent,
				 0 );
/*        printf( "%d %ld\n", keycode, p->keysym ); */
      if ((p->keysym == XK_Q) || (p->keysym == XK_q)) {
	return( EV_QUIT );
      }
      break;
    case MotionNotify:
      return( EV_MOTION );
      break;
    }
  }
  return( EV_NOTHING );
}

int32 updateBackground( GraphicsData *gr, GridInfo *gi )
{
  int32 i, x, y, w, h;

  if (gr->isBackground) {
    for (i = 0; i < (gi->nx * gi->ny); i++) {
      if (gi->mask[i] && (!gi->mask0[i])) {
	x = gi->x[i%gi->nx];
	w = gi->w[i%gi->nx];
	y = gi->y[i/gi->nx];
	h = gi->h[i/gi->nx];
	/* Update background. */
	gr->back = XGetSubImage( gr->thedisplay, gr->thewindow,
				 x, y, w, h, AllPlanes, ZPixmap,
				 gr->back, x, y );
      }
    }
  }
  return( 0 );
}

int32 insertImage( uchar *p1, GraphicsData *gr, GridInfo *gi )
{
  char *sp, *dp, *bp;
  int32 i, j, x, y, w, h;
  int32 na;

  na = 0;
  for (i = 0; i < (gi->nx * gi->ny); i++) {
    if (gi->mask[i]) {
      na++;
      x = gi->x[i%gi->nx];
      w = gi->w[i%gi->nx];
      y = gi->y[i/gi->nx];
      h = gi->h[i/gi->nx];
/*        printf( "%d %d %d %d\n", x, y, w, h ); */
      sp = (char *) p1 + gr->w * y + x;
      dp = (char *) gr->image + gr->bytesPerPixel * (gr->w * y + x);
      /* Draw it. */
      if (gr->isBackground) {
	bp = (char *) gr->back->data + gr->bytesPerPixel * (gr->w * y + x);
	for (j = 0; j < h; j++) {
	  memcpy( dp, bp, gr->bytesPerPixel * w * sizeof( char ) );
	  gr->convert( sp, dp, gr->pal, w );
	  sp += gr->w;
	  dp += gr->bytesPerPixel * gr->w;
	  bp += gr->bytesPerPixel * gr->w;
	}
      } else {
	for (j = 0; j < h; j++) {
	  gr->convert( sp, dp, gr->pal, w );
	  sp += gr->w;
	  dp += gr->bytesPerPixel * gr->w;
	}
      }
      XPutImage( gr->thedisplay, gr->thewindow, gr->thecontext, gr->ximg,
	       x, y, x, y, w, h );
    }
  }

/*    printf( "%d ", na ); */
/*    XPutImage( gr->thedisplay, gr->thewindow, gr->thecontext, gr->ximg, */
/*  	     0, 0, gr->xr + (gr->wr - gr->w) / 2, */
/*  	     gr->yr + (gr->hr - gr->h) / 2, gr->w, gr->h ); */
/*    XPutImage( gr->thedisplay, gr->thewindow, gr->thecontext, gr->ximg, */
/*  	     0, 0, 0, 0, gr->w, gr->h ); */
  /*  XFlush( gr->thedisplay );*/
  return( 0 );
}

void putString( int32 x, int32 y, char *text, int color, GraphicsData *p )
{
  XSetForeground(p->thedisplay,p->gcInfo,color);
  XDrawString( p->thedisplay, p->pixInfo, p->gcInfo,
               x,
               y, text, strlen( text ) );
}

void putStats( int32 x, int32 y, char *s1, char *s2, GraphicsData *p )
{
  XSetForeground(p->thedisplay,p->gcInfo,p->blackColor);
  XFillRectangle( p->thedisplay, p->pixInfo, p->gcInfo,
		  0, 0, 80, 40 );
  putString( 5, 20, s1, p->whiteColor, p );
  putString( 5, 35, s2, p->whiteColor, p );
  XCopyArea( p->thedisplay, p->pixInfo, p->thewindow, p->thecontext, 0, 0,
	     80, 40, x, y );
}
