#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
/*  #include <sys/time.h> */

#include "memutils.h"
#include "fire-sparse.h"
#include "graphics.h"
#include "handle.h"
#include "timer.h"

volatile sig_atomic_t awaken = 0; 
struct itimerval itim;

void noHandler( int sig )
{

/*    printf( "bubu\n" ); */
  signal( SIGALRM, noHandler );
  if (!awaken) {
    itim.it_interval.tv_sec = 0;
    itim.it_interval.tv_usec = 0;
    itim.it_value.tv_sec = 0;
    itim.it_value.tv_usec = 10000;
    setitimer( ITIMER_REAL, &itim, 0 );
  }
}

void swapPointers( void **p1, void **p2 )
{
  void *p;

  p = *p1;
  *p1 = *p2;
  *p2 = p;
}

void updateMask( GridInfo *gi, FSparseData *d, FParams *par )
{
  int32 i, of, x, y;

  memset( gi->mask, 0, gi->nx * gi->ny * sizeof( uchar ) );
  for (i = 0; i < d->nFP; i++) {
    of = d->listFP[i].offset;
    x = gi->xmap[of%par->fireW];
    y = gi->ymap[of/par->fireW];
    gi->mask[x+y*gi->nx] = 1;
  }
/*    gi->mask[gi->nx*gi->ny-1] = 1; */
/*    gi->mask[gi->nx*gi->ny-2] = 1; */
}

/*
 * m1 = m2 | m3.
 */
void combineMask( uchar *m1, uchar *m2, uchar *m3, int32 n )
{
  int32 i;

  for (i = 0; i < n; i++) {
    m1[i] = m2[i] | m3[i];
  }
}

int main( int argc, char **argv )
{
  Filter f;
  FParams par;
  FSparseData *d1, *d2;
  GraphicsData gr;
  Timer t, total;
  int32 ret;
  char txtbuf[256]; /* !!! */
  char txtbuf2[256]; /* !!! */
  int frameTag;
  float isTime;
  time_t user, sys;
  clock_t tot, tot0; 
  float32 t1f, t1f0, fps;
/*    int x, y; */
  int i, j, justOne, aux;
  GridInfo gi;
  int32 maxGridSize = 50;
  int32 dd;

  if (fsLoadParameters( &par )) {
    fprintf( stderr, "Error during loading input!\n" );
    exit(1);
  }

  if (par.fPS > 100) {
    par.fPS = 100;
    printf( "Reducing FPS to %d\n", par.fPS );
  }
  fps = par.fPS;

  gr.inRootWindow = 0;
  gr.isBackground = 0;
  gr.colormapFile = 0;
  if ((argc == 2) || (argc == 3)) {
    if (!strcmp( argv[1], "-root" )) {
      gr.inRootWindow = 1;
    } else if (!strcmp( argv[1], "-back" )) {
	gr.inRootWindow = 1;
	gr.isBackground = 1;
    }
  }
  if ((argc == 2) && (!gr.inRootWindow)) {
    gr.colormapFile = argv[1];
  } else if ((argc == 3)) {
    gr.colormapFile = argv[2];
  }
  initGraphics( &gr, "fire", &par.fireW, &par.fireH );

  signal( SIGALRM, noHandler );

  gi.nx = (par.fireW / maxGridSize);
  if (par.fireW % maxGridSize) gi.nx++;
  dd = maxGridSize;
  gi.x = (int32 *) allocMem( gi.nx * sizeof( int32 ) );
  gi.w = (int32 *) allocMem( gi.nx * sizeof( int32 ) );
  gi.xmap = (int32 *) allocMem( par.fireW * sizeof( int32 ) );
  for (i = 0; i < gi.nx; i++) {
    gi.w[i] = dd;
    gi.x[i] = i * dd;
  }
  gi.w[gi.nx-1] = par.fireW - gi.x[gi.nx-1];
  for (i = 0; i < gi.nx; i++) {
    for (j = 0; j < gi.w[i]; j++) {
      gi.xmap[gi.x[i]+j] = i;
    }
  }

  gi.ny = (par.fireH / maxGridSize);
  if (par.fireH % maxGridSize) gi.ny++;
  dd = maxGridSize;
  gi.y = (int32 *) allocMem( gi.ny * sizeof( int32 ) );
  gi.h = (int32 *) allocMem( gi.ny * sizeof( int32 ) );
  gi.ymap = (int32 *) allocMem( par.fireH * sizeof( int32 ) );
  for (i = 0; i < gi.ny; i++) {
    gi.h[i] = dd;
    gi.y[i] = i * dd;
  }
  gi.h[gi.ny-1] = par.fireH - gi.y[gi.ny-1];
  for (i = 0; i < gi.ny; i++) {
    for (j = 0; j < gi.h[i]; j++) {
      gi.ymap[gi.y[i]+j] = i;
    }
  }

  printf( "Grid size: %d %d\n", gi.nx, gi.ny );

/*    for (i = 0; i < gi.nx; i++) { */
/*      printf( "%d %d\n", gi.x[i], gi.w[i] ); */
/*    } */
/*    for (i = 0; i < gi.ny; i++) { */
/*      printf( "%d %d\n", gi.y[i], gi.h[i] ); */
/*    } */

  gi.mask = (uchar *) allocMem( gi.nx * gi.ny * sizeof( uchar ) );
  gi.mask0 = (uchar *) allocMem( gi.nx * gi.ny * sizeof( uchar ) );

  d1 = (FSparseData *) allocMem( sizeof(FSparseData) );
  d2 = (FSparseData *) allocMem( sizeof(FSparseData) );
  fsCreateData( &par, d1 );
  fsCreateData( &par, d2 );
  fsClearData( &par, d1 );
  fsClearData( &par, d2 );

/*    for (x = 140, y = 140; x < 150; x++, y++) { */
/*      for (i = x - 10; i <= x + 10; i++) { */
/*        addFPixel( d1->bufFP, d1->listFP, &d1->nFP, par.maxFPixels, */
/*  		 i + y * par.fireW, par.maxPower );	     */
/*        addFPixel( d1->bufFP, d1->listFP, &d1->nFP, par.maxFPixels, */
/*  		 i + (y + 1) * par.fireW, par.maxPower );	     */
/*        addFPixel( d1->bufFP, d1->listFP, &d1->nFP, par.maxFPixels, */
/*  		 i + (y - 1) * par.fireW, par.maxPower );	     */
/*      }						     */
/*      for (j = y - 10; j <= y + 10; j++) {		     */
/*        addFPixel( d1->bufFP, d1->listFP, &d1->nFP, par.maxFPixels, */
/*  		 x + j * par.fireW, par.maxPower );	     */
/*        addFPixel( d1->bufFP, d1->listFP, &d1->nFP, par.maxFPixels, */
/*  		 x + 1 + j * par.fireW, par.maxPower );	     */
/*        addFPixel( d1->bufFP, d1->listFP, &d1->nFP, par.maxFPixels, */
/*  		 x - 1 + j * par.fireW, par.maxPower ); */
/*      } */
/*    } */

  f.init = FPointer &fsInit;
  f.action = FPointer &fsAction;
  f.clear = FPointer &fsClear;
  f.type = FSFilter;
  f.data = 0;
  
  t1f = t1f0 = (float) CLK_TCK / par.fPS;
  printf( "Timing: %d %d %f\n", CLOCKS_PER_SEC, CLK_TCK, t1f );
  timerStart( &total );
  timerStart( &t );
  f.init( &f, &par );
  frameTag = 0;
  tot0 = 0;
  sprintf( txtbuf, "undetermined" );
  while (1) {
    /* One CA iteration. */
    f.action( &f, d1, d2 );
    /* Make 'd1' the current pointer. */
    swapPointers( (void **) &d1, (void **) &d2 );

    /* Sets new 'gi.mask' according to 'd1'. */
    updateMask( &gi, d1, &par );

    updateBackground( &gr, &gi );

    /* Clears 'd2' according to previous mask 'gi.mask0'. */
    fsClearDataFast( &par, d2, &gi );
    /*
     * Combine both grid masks for drawing. This is required to have both
     * old cells cleared and new drawn in the first frame they appear in.
     * Old way was to update drawing mask after drawing, so that old grid
     * cells were cleared, but there was a one frame delay after what was
     * computed was actually drawn -> there were rectangular 'cuts'
     * in the newest source when crossing grid borders,
     * since the new grid cell was drawn
     * only after next frame, which could take a long time...
     */
    combineMask( gi.mask0, gi.mask0, gi.mask, gi.nx * gi.ny );
    /* 'mask' is now combination of new and previous, 'mask0' is new only. */
    swapPointers( (void **) &(gi.mask), (void **) &(gi.mask0) );

    /* Draw it all. */
    insertImage( d1->bufFP, &gr, &gi );
    sprintf( txtbuf2, "%d", d1->nFP );
    putStats( gr.wr - 80, gr.hr - 40, txtbuf, txtbuf2, &gr );

    /* Decrease frame rate when there is no stuff to draw here. */
    if ((d1->nFP == 0)) t1f = t1f0 * 10;
    else t1f = t1f0;

    /*
     * Check events and handle timing and sleeping.
     */
    isTime = 0.0;
    while (1) {
      justOne = (fps < (0.7 * par.fPS));
      aux = 1;
      while ((ret = processEvents( &gr )) == EV_MOTION) {
	if ((ret != EV_NOTHING) && aux) {
/*  	  if (d1->nFP < 12000) */
	    ret = handleEvents( d1, &gr, &par, ret );
	}
	if (justOne) aux = 0;
      }
      if (ret == EV_QUIT) goto main_end;
      timerCurrentStatus( &t, &tot, &user, &sys );
      /*      printf( "%d %d %d\n", tot, user, sys );*/
      /*      printf( "%d %d\n", tot, t1f );*/
/*        printf( "%f\n", tot - 0.8 * t1f ); */
      isTime =  0.8 * t1f - (float) tot;
      if (isTime <= 0.0) {
	timerStart( &t );
	break;
      } else {
	if ((d1->nFP == 0)) {
	  sprintf( txtbuf, "sleeping" );
/*  	  if (isTime < 1.0) isTime = 1.0; */
	  itim.it_interval.tv_sec = 0;
	  itim.it_interval.tv_usec = 0;
	  itim.it_value.tv_sec = 0;
/*  	  itim.it_value.tv_usec = (1000000.0 * (isTime / (float) CLK_TCK)); */
	  itim.it_value.tv_usec = 150000;
	  awaken = 0;
	  setitimer( ITIMER_REAL, &itim, 0 );
	  pause();
	  awaken = 1;
	}
      }
    }

    /* Update FPS string. */
    timerCurrentStatus( &total, &tot, &user, &sys );
    if ((tot > tot0 + CLK_TCK) && (d1->nFP > 0)) {
      fps = (float) frameTag / ((float) (tot - tot0) / CLK_TCK);
      sprintf( txtbuf, "%f", fps );
      if (tot < t1f) strcat( txtbuf, "+" );
/*        printf( "%s\n", txtbuf ); */
      tot0 = tot;
      frameTag = 0;
    } 
    frameTag++;
  }

 main_end:

  f.clear( &f );

  clearGraphics( &gr );

  fsDestroyData( d1 );
  fsDestroyData( d2 );
  freeMem( d1 );
  freeMem( d2 );
  freeMem( gi.x );
  freeMem( gi.y );
  freeMem( gi.w );
  freeMem( gi.h );
  freeMem( gi.xmap );
  freeMem( gi.ymap );
  freeMem( gi.mask );
  freeMem( gi.mask0 );

  return( 0 );
}
